/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	Declaration of atomic manipulation functions.
 *
 * Environment:
 *	All
 *
 * $Revision: 1.1 $
 */

#ifndef _CL_ATOMIC_H_
#define _CL_ATOMIC_H_

#include <complib/cl_atomic_osd.h>

/****h* Component Library/Atomic Operations
* NAME
*	Atomic Operations
*
* DESCRIPTION
*	The Atomic Operations functions allow callers to operate on
*	32-bit signed integers in an atomic fashion.
*********/

#ifdef __cplusplus
extern "C" {
#endif

/****f* Component Library: Atomic Operations/cl_atomic_inc
* NAME
*	cl_atomic_inc
*
* DESCRIPTION
*	The cl_atomic_inc function atomically increments a 32-bit signed
*	integer and returns the incremented value.
*
* SYNOPSIS
*/
	int32_t cl_atomic_inc(IN atomic32_t * const p_value);
/*
* PARAMETERS
*	p_value
*		[in] Pointer to a 32-bit integer to increment.
*
* RETURN VALUE
*	Returns the incremented value pointed to by p_value.
*
* NOTES
*	The provided value is incremented and its value returned in one atomic
*	operation.
*
*	cl_atomic_inc maintains data consistency without requiring additional
*	synchronization mechanisms in multi-threaded environments.
*
* SEE ALSO
*	Atomic Operations, cl_atomic_dec, cl_atomic_add, cl_atomic_sub,
*	cl_atomic_xchg, cl_atomic_comp_xchg
*********/

/****f* Component Library: Atomic Operations/cl_atomic_dec
* NAME
*	cl_atomic_dec
*
* DESCRIPTION
*	The cl_atomic_dec function atomically decrements a 32-bit signed
*	integer and returns the decremented value.
*
* SYNOPSIS
*/
	 int32_t cl_atomic_dec(IN atomic32_t * const p_value);
/*
* PARAMETERS
*	p_value
*		[in] Pointer to a 32-bit integer to decrement.
*
* RETURN VALUE
*	Returns the decremented value pointed to by p_value.
*
* NOTES
*	The provided value is decremented and its value returned in one atomic
*	operation.
*
*	cl_atomic_dec maintains data consistency without requiring additional
*	synchronization mechanisms in multi-threaded environments.
*
* SEE ALSO
*	Atomic Operations, cl_atomic_inc, cl_atomic_add, cl_atomic_sub,
*	cl_atomic_xchg, cl_atomic_comp_xchg
*********/

/****f* Component Library: Atomic Operations/cl_atomic_add
* NAME
*	cl_atomic_add
*
* DESCRIPTION
*	The cl_atomic_add function atomically adds a value to a
*	32-bit signed integer and returns the resulting value.
*
* SYNOPSIS
*/
	 int32_t
	    cl_atomic_add(IN atomic32_t * const p_value,
			  IN const int32_t increment);
/*
* PARAMETERS
*	p_value
*		[in] Pointer to a 32-bit integer that will be added to.
*
*	increment
*		[in] Value by which to increment the integer pointed to by p_value.
*
* RETURN VALUE
*	Returns the value pointed to by p_value after the addition.
*
* NOTES
*	The provided increment is added to the value and the result returned in
*	one atomic operation.
*
*	cl_atomic_add maintains data consistency without requiring additional
*	synchronization mechanisms in multi-threaded environments.
*
* SEE ALSO
*	Atomic Operations, cl_atomic_inc, cl_atomic_dec, cl_atomic_sub,
*	cl_atomic_xchg, cl_atomic_comp_xchg
*********/

/****f* Component Library: Atomic Operations/cl_atomic_sub
* NAME
*	cl_atomic_sub
*
* DESCRIPTION
*	The cl_atomic_sub function atomically subtracts a value from a
*	32-bit signed integer and returns the resulting value.
*
* SYNOPSIS
*/
	 int32_t
	    cl_atomic_sub(IN atomic32_t * const p_value,
			  IN const int32_t decrement);
/*
* PARAMETERS
*	p_value
*		[in] Pointer to a 32-bit integer that will be subtracted from.
*
*	decrement
*		[in] Value by which to decrement the integer pointed to by p_value.
*
* RETURN VALUE
*	Returns the value pointed to by p_value after the subtraction.
*
* NOTES
*	The provided decrement is subtracted from the value and the result
*	returned in one atomic operation.
*
*	cl_atomic_sub maintains data consistency without requiring additional
*	synchronization mechanisms in multi-threaded environments.
*
* SEE ALSO
*	Atomic Operations, cl_atomic_inc, cl_atomic_dec, cl_atomic_add,
*	cl_atomic_xchg, cl_atomic_comp_xchg
*********/

/****f* Component Library: Atomic Operations/cl_atomic_xchg
* NAME
*	cl_atomic_xchg
*
* DESCRIPTION
*	The cl_atomic_xchg function atomically sets a value of a
*	32-bit signed integer and returns the initial value.
*
* SYNOPSIS
*/
	 int32_t
	    cl_atomic_xchg(IN atomic32_t * const p_value,
			   IN const int32_t new_value);
/*
* PARAMETERS
*	p_value
*		[in] Pointer to a 32-bit integer to exchange with new_value.
*
*	new_value
*		[in] Value to assign.
*
* RETURN VALUE
*	Returns the initial value pointed to by p_value.
*
* NOTES
*	The provided value is exchanged with new_value and its initial value
*	returned in one atomic operation.
*
*	cl_atomic_xchg maintains data consistency without requiring additional
*	synchronization mechanisms in multi-threaded environments.
*
* SEE ALSO
*	Atomic Operations, cl_atomic_inc, cl_atomic_dec, cl_atomic_add,
*	cl_atomic_sub, cl_atomic_comp_xchg
*********/

/****f* Component Library: Atomic Operations/cl_atomic_comp_xchg
* NAME
*	cl_atomic_comp_xchg
*
* DESCRIPTION
*	The cl_atomic_comp_xchg function atomically compares a 32-bit signed
*	integer to a desired value, sets that integer to the
*	specified value if equal, and returns the initial value.
*
* SYNOPSIS
*/
	 int32_t
	    cl_atomic_comp_xchg(IN atomic32_t * const p_value,
				IN const int32_t compare,
				IN const int32_t new_value);
/*
* PARAMETERS
*	p_value
*		[in] Pointer to a 32-bit integer to exchange with new_value.
*
*	compare
*		[in] Value to compare to the value pointed to by p_value.
*
*	new_value
*		[in] Value to assign if the value pointed to by p_value is equal to
*		the value specified by the compare parameter.
*
* RETURN VALUE
*	Returns the initial value of the variable pointed to by p_value.
*
* NOTES
*	The value pointed to by p_value is compared to the value specified by the
*	compare parameter.  If the two values are equal, the p_value variable is
*	set to new_value.  The initial value pointed to by p_value is returned.
*
*	cl_atomic_comp_xchg maintains data consistency without requiring additional
*	synchronization mechanisms in multi-threaded environments.
*
* SEE ALSO
*	Atomic Operations, cl_atomic_inc, cl_atomic_dec, cl_atomic_add,
*	cl_atomic_sub, cl_atomic_xchg
*********/

#ifdef __cplusplus
}				/* extern "C" */
#endif
#endif				/* _CL_ATOMIC_H_ */
