/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	Declaration of mutex object.
 *
 * Environment:
 *	All
 *
 * $Revision: 1.1 $
 */

#ifndef _CL_MUTEX_H_
#define _CL_MUTEX_H_

#include <complib/cl_mutex_osd.h>

/****h* complib/Mutex
* NAME
*	Mutex
*
* DESCRIPTION
*	Mutex provides synchronization between threads for exclusive access to
*	a resource.
*
*	The Mutex functions manipulate a cl_mutex_t structure which should
*	be treated as opaque and should be manipulated only through the provided
*	functions.
*
* SEE ALSO
*	Structures:
*		cl_mutex_t
*
*	Initialization:
*		cl_mutex_construct, cl_mutex_init, cl_mutex_destroy
*
*	Manipulation
*		cl_mutex_acquire, cl_mutex_release
*********/

#ifdef __cplusplus
extern "C" {
#endif

/****f* Component Library: Mutex/cl_mutex_construct
* NAME
*	cl_mutex_construct
*
* DESCRIPTION
*	The cl_mutex_construct function initializes the state of a
*	mutex.
*
* SYNOPSIS
*/
	void
	 cl_mutex_construct(IN cl_mutex_t * const p_mutex);
/*
* PARAMETERS
*	p_mutex
*		[in] Pointer to a mutex structure whose state to initialize.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Allows calling cl_semphore_destroy without first calling
*	cl_mutex_init.
*
*	Calling cl_mutex_construct is a prerequisite to calling any other
*	mutex function except cl_mutex_init.
*
* SEE ALSO
*	Mutex, cl_semphore_init cl_mutex_destroy
*********/

/****f* Component Library: Mutex/cl_mutex_init
* NAME
*	cl_mutex_init
*
* DESCRIPTION
*	The cl_mutex_init function initializes a mutex for use.
*
* SYNOPSIS
*/
	 cl_status_t cl_mutex_init(IN cl_mutex_t * const p_mutex);
/*
* PARAMETERS
*	p_mutex
*		[in] Pointer to a mutex structure to initialize.
*
* RETURN VALUES
*	CL_SUCCESS if initialization succeeded.
*
*	CL_ERROR if initialization failed. Callers should call
*	cl_mutex_destroy to clean up any resources allocated during
*	initialization.
*
* NOTES
*	Initializes the mutex structure. Allows calling cl_mutex_aquire
*	and cl_mutex_release. The cl_mutex is always created in the unlocked state.
*
* SEE ALSO
*	Mutex, cl_mutex_construct, cl_mutex_destroy,
*	cl_mutex_acquire, cl_mutex_release
*********/

/****f* Component Library: Mutex/cl_mutex_destroy
* NAME
*	cl_mutex_destroy
*
* DESCRIPTION
*	The cl_mutex_destroy function performs all necessary cleanup of a
*	mutex.
*
* SYNOPSIS
*/
	void
	 cl_mutex_destroy(IN cl_mutex_t * const p_mutex);
/*
* PARAMETERS
*	p_mutex
*		[in] Pointer to a mutex structure to destroy.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Performs any necessary cleanup of a mutex. This function must only
*	be called if either cl_mutex_construct or cl_mutex_init has been
*	called.
*
* SEE ALSO
*	Mutex, cl_mutex_construct, cl_mutex_init
*********/

/****f* Component Library: Mutex/cl_mutex_acquire
* NAME
*	cl_mutex_acquire
*
* DESCRIPTION
*	The cl_mutex_acquire function acquires a mutex.
*
* SYNOPSIS
*/
	void
	 cl_mutex_acquire(IN cl_mutex_t * const p_mutex);
/*
* PARAMETERS
*	p_mutex
*		[in] Pointer to a mutex structure to acquire.
*
* RETURN VALUE
*	This function does not return a value.
*
* SEE ALSO
*	Mutex, cl_mutex_release
*********/

/****f* Component Library: Mutex/cl_mutex_release
* NAME
*	cl_mutex_release
*
* DESCRIPTION
*	The cl_mutex_release function releases a mutex object.
*
* SYNOPSIS
*/
	void
	 cl_mutex_release(IN cl_mutex_t * const p_mutex);
/*
* PARAMETERS
*	p_mutex
*		[in] Pointer to a mutex structure to release.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Releases a mutex after a call to cl_mutex_acquire.
*
* SEE ALSO
*	Mutex, cl_mutex_acquire
*********/

#ifdef __cplusplus
}				/* extern "C" */
#endif
#endif				/* _CL_MUTEX_H_ */
