/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	This module defines the initialization and cleanup sequences for the verbs
 *	provider driver.
 *
 * Environment:
 *	Linux Kernel Mode
 *
 * $Revision: 1.1 $
 */

#include <complib/cl_types.h>
#include <complib/cl_device.h>
#include <complib/cl_debug.h>
#include <complib/cl_syshelper.h>
#include <complib/cl_memory.h>
#include <linux/proc_fs.h>
#include <complib/cl_verinfo.h>

#include <linux/module.h>
#include <linux/kernel.h>

MODULE_DESCRIPTION("Linux Component Library");
MODULE_LICENSE("Dual BSD/GPL");

/*
 *
 * All CompLib specific command line parameters
 */
extern int log2cons;
MODULE_PARM(log2cons, "i");
MODULE_PARM_DESC(log2cons,
		 "Control debug output on console, or log file (0-1)");

int mem_track = 1;
MODULE_PARM(mem_track, "i");
MODULE_PARM_DESC(mem_track, "Enable memory allocation tracking");

/*
 * Driver Globals
 */
DRV_ID_STR("complib");

struct proc_dir_entry *p_cl_proc_dir = NULL;

extern
 cl_status_t __cl_sys_callback_init(void);

extern
    void
 __cl_sys_callback_destroy(void);

void
cl_create_proc(void)
{
	extern void create_cl_proc_mem(struct proc_dir_entry *p_dir);

	p_cl_proc_dir = proc_mkdir("complib", NULL);

	if (!p_cl_proc_dir) {
		cl_msg_out("cl_create_proc:Unable to create proc dir\n");
		return;
	}

	/*
	 *  Call the memtrack proc init code
	 */
	create_cl_proc_mem(p_cl_proc_dir);
}

void
cl_destroy_proc(void)
{
	extern void destroy_cl_proc_mem(void);

	if (p_cl_proc_dir) {
		destroy_cl_proc_mem();
		remove_proc_entry("complib", NULL);
	}
}

/*
 * Initialization routine - called when the driver is loaded
 */

int __init
complib_init(void)
{
	cl_status_t status = CL_SUCCESS;

	/* Startup all services */
	PRINT_MOD_INFO("Linux Component Library");
	__cl_mem_track(TRUE);

	/* System Device Framework Support */
	status = __cl_dev_frmwk_init();
	if (status != CL_SUCCESS) {
		cl_msg_out("Failed to startup system device \n");
		goto complib_load_fail;
	}

	/*
	 * Next, startup the syshelper framework.
	 * Note: This "__cl_syshelper_init()" can be called
	 * after __cl_dev_frmwk_init() is called because it uses
	 * the device framework. Please do not change the order.
	 */

	status = __cl_syshelper_init();
	if (status != CL_SUCCESS) {
		cl_msg_out(" Failed to startup system helper \n");
		goto complib_load_fail;
	}

	status = __cl_sys_callback_init();
	if (status != CL_SUCCESS) {
		cl_msg_out(" Failed to startup system callback \n");
	}

	/*
	 * Create Proc file interfaces
	 */
	cl_create_proc();
	return CL_SUCCESS;

      complib_load_fail:

	__cl_syshelper_exit();
	__cl_dev_frmwk_destroy();
	__cl_mem_track(FALSE);

	return status;
}

/*
 * Cleanup routine - called when the driver is unloaded
 */

void __exit
complib_cleanup(void)
{
	/* Stop all services */

	/*
	 * As pointed out in complib_init (above), there
	 * is a well defined ordering to these functions.
	 * Please do not change it.
	 */

	__cl_sys_callback_destroy();

	__cl_syshelper_exit();
	__cl_dev_frmwk_destroy();

	cl_destroy_proc();

	__cl_mem_track(FALSE);
}

module_init(complib_init);
module_exit(complib_cleanup);
