/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	Spinlock Implementations for Linux kernel modules.
 *
 * Environment
 *	Linux Kernel Mode
 *
 * $Revision: 1.1 $
 */

#include <complib/cl_spinlock.h>
#include <complib/cl_memory.h>
#include <complib/cl_debug.h>
#include <linux/sched.h>

/* Construct a spin lock */
void
cl_spinlock_construct(IN cl_spinlock_t * const p_spinlock)
{
	CL_ASSERT(p_spinlock);

	cl_memclr(p_spinlock, sizeof (cl_spinlock_t));
	p_spinlock->state = CL_UNINITIALIZED;
}

/* Initialize a spin lock */
cl_status_t
cl_spinlock_init(IN cl_spinlock_t * const p_spinlock)
{
	CL_ASSERT(p_spinlock);

	/* Call the constructor if it hasn't yet. */
	cl_spinlock_construct(p_spinlock);

	spin_lock_init(&p_spinlock->lock);
	p_spinlock->flags = 0;
	p_spinlock->state = CL_INITIALIZED;
	p_spinlock->level = SPIN_LVL_INVALID;

#ifdef _DEBUG_
	p_spinlock->locked = FALSE;
	p_spinlock->owner = NULL;
	p_spinlock->cpuid = -1;
#endif
	return (CL_SUCCESS);
}

/* Destroy a spin lock */
void
cl_spinlock_destroy(IN cl_spinlock_t * const p_spinlock)
{
	CL_ASSERT(p_spinlock);
	CL_ASSERT(cl_is_state_valid(p_spinlock->state));

	p_spinlock->level = SPIN_LVL_INVALID;
#ifdef _DEBUG_
	if (p_spinlock->locked) {
		cl_dbg_out
		    ("cl_spinlock_destroy():Lock %p: locked : 0x%x owner : "
		     "(%s) 0x%p cpuid : 0x%x\n", p_spinlock, p_spinlock->locked,
		     p_spinlock->owner->comm, p_spinlock->owner,
		     p_spinlock->cpuid);
		CL_ASSERT(0);
	}
	p_spinlock->state = CL_DESTROYED;
#endif
}

#ifdef _DEBUG_
static void
__cl_spinlock_acq_dbg(IN cl_spinlock_t * const p_spinlock)
{
	CL_ASSERT(p_spinlock->locked == FALSE);
	p_spinlock->locked = TRUE;
	p_spinlock->owner = current;
	p_spinlock->cpuid = smp_processor_id();
}

static void
__cl_spinlock_rel_dbg(IN cl_spinlock_t * const p_spinlock)
{
	if ((p_spinlock->locked != TRUE) ||
	    (p_spinlock->owner != current) ||
	    (p_spinlock->cpuid != smp_processor_id())) {
		if (p_spinlock->owner) {
			cl_dbg_out
			    ("cl_spinlock_release():Lock %p: locked : 0x%x owner : "
			     "(%s) 0x%p cpuid : 0x%x\n", p_spinlock,
			     p_spinlock->locked, p_spinlock->owner->comm,
			     p_spinlock->owner, p_spinlock->cpuid);
		} else {
			cl_dbg_out
			    ("cl_spinlock_release():Lock %p: locked : 0x%x cpuid : "
			     "0x%x\n", p_spinlock, p_spinlock->locked,
			     p_spinlock->cpuid);
		}

		cl_dbg_out("current : (%s) 0x%p cur cpuid : 0x%x\n",
			   current->comm, current, smp_processor_id());

		CL_ASSERT(((p_spinlock->locked == TRUE) &&
			   (p_spinlock->owner == current)));
	}
	p_spinlock->locked = FALSE;
	p_spinlock->owner = NULL;
	p_spinlock->cpuid = -1;
}
#else
#define __cl_spinlock_acq_dbg(x)
#define __cl_spinlock_rel_dbg(x)
#endif

/* Acquire a spin lock */
void
cl_spinlock_acquire_irq(IN cl_spinlock_t * const p_spinlock)
{
	unsigned long flags;

	CL_ASSERT(p_spinlock);
	CL_ASSERT(p_spinlock->state == CL_INITIALIZED);

	spin_lock_irqsave(&p_spinlock->lock, flags);
	p_spinlock->flags = flags;
	p_spinlock->level = SPIN_LVL_INTERRUPT;

	__cl_spinlock_acq_dbg(p_spinlock);
}

void
cl_spinlock_acquire(IN cl_spinlock_t * const p_spinlock)
{
	CL_ASSERT(p_spinlock);
	CL_ASSERT(p_spinlock->state == CL_INITIALIZED);

	spin_lock_bh(&p_spinlock->lock);
	p_spinlock->level = SPIN_LVL_TASKLET;

	__cl_spinlock_acq_dbg(p_spinlock);
}

/* Release a spin lock */
void
cl_spinlock_release_irq(IN cl_spinlock_t * const p_spinlock)
{
	CL_ASSERT(p_spinlock);
	CL_ASSERT(p_spinlock->state == CL_INITIALIZED);

	__cl_spinlock_rel_dbg(p_spinlock);
	p_spinlock->level = SPIN_LVL_INVALID;
	spin_unlock_irqrestore(&p_spinlock->lock, p_spinlock->flags);
}

void
cl_spinlock_release(IN cl_spinlock_t * const p_spinlock)
{
	CL_ASSERT(p_spinlock);
	CL_ASSERT(p_spinlock->state == CL_INITIALIZED);

	__cl_spinlock_rel_dbg(p_spinlock);
	p_spinlock->level = SPIN_LVL_INVALID;
	spin_unlock_bh(&p_spinlock->lock);
}
