/*++
"This software program is available to you under a choice of one of two 
licenses.  You may choose to be licensed under either the GNU General Public 
License (GPL) Version 2, June 1991, available at 
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License, 
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing 
to grant a license for the software entitled InfiniBand(tm) System Software 
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License: 

"Licensed Patents" means patent claims licensable by Intel Corporation which 
are necessarily infringed by the use or sale of the Software alone or when 
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to 
any operating system available under the GNU Public License version 2.0 or 
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved. 

The license is provided to Recipient and Recipient's Licensees under the 
following terms.  

Redistribution and use in source and binary forms of the Software, with or 
without modification, are permitted provided that the following conditions are 
met: 
Redistributions of source code of the Software may retain the above copyright 
notice, this list of conditions and the following disclaimer. 

Redistributions in binary form of the Software may reproduce the above 
copyright notice, this list of conditions and the following disclaimer in the 
documentation and/or other materials provided with the distribution. 

Neither the name of Intel Corporation nor the names of its contributors shall 
be used to endorse or promote products derived from this Software without 
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide, 
royalty-free patent license under Licensed Patents to make, use, sell, offer 
to sell, import and otherwise transfer the Software, if any, in source code 
and object code form. This license shall include changes to the Software that 
are error corrections or other minor changes to the Software that do not add 
functionality or features when the Software is incorporated in any version of 
a operating system that has been distributed under the GNU General Public 
License 2.0 or later.  This patent license shall apply to the combination of 
the Software and any operating system licensed under the GNU Public License 
version 2.0 or later if, at the time Intel provides the Software to Recipient, 
such addition of the Software to the then publicly available versions of such 
operating system available under the GNU Public License version 2.0 or later 
(whether in gold, beta or alpha form) causes such combination to be covered by 
the Licensed Patents. The patent license shall not apply to any other 
combinations which include the Software. No hardware per se is licensed 
hereunder. 
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY 
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	Implementation of thread abstractions. Create, destroy, signal a thread
 *	to perform work etc.
 *
 * Environment
 *	Linux Kernel Mode
 * 
 * $Revision: 1.1 $
 */

#include <complib/cl_thread.h>
#include <complib/cl_memory.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/signal.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#define SHUTDOWN_SIGS	(sigmask(SIGKILL) | sigmask(SIGINT) | sigmask(SIGTERM))

/* 
 * The function that kernel_thread() calls when it forks off a
 * child.
 */
int
__cl_kernel_system_thread(IN void *const arg)
{
	cl_thread_t *p_thread = (cl_thread_t *) arg;

	CL_ASSERT(p_thread);

	lock_kernel();

	/* Fix the associations with the parent thread.
	 * This only works for kernels > 2.2.18 
	 */
#ifndef LINUX_VERSION_CODE
#error You must include linux/version.h somewhere before here
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	daemonize(p_thread->name);
#else
	siginitsetinv(&current->blocked, SHUTDOWN_SIGS);

	daemonize();
	sprintf(current->comm, p_thread->name);
#endif

	unlock_kernel();

	p_thread->osd.task = current;

	/* Invoke the user's callback function */
	p_thread->pfn_callback((void *) p_thread->context);

	/* If the user thread returns, we terminate the thread */
	p_thread->osd.task = NULL;

	/* People calling destroy should be allowed to go now */
	cl_event_signal(&p_thread->osd.kill_event);

	return (0);
}

/* Indicate that there was an intent to use a thread */
void
cl_thread_construct(IN cl_thread_t * const p_thread)
{
	cl_memclr(p_thread, sizeof (cl_thread_t));
	cl_event_construct(&p_thread->osd.kill_event);

	p_thread->osd.state = CL_UNINITIALIZED;
}

/* 
 * Create a new thread. This thread will execute supplied "Callback"
 * with the user supplied "Context"
 */
cl_status_t
cl_thread_init(IN cl_thread_t * const p_thread,
	       IN cl_pfn_thread_callback_t pfn_callback,
	       IN const void *const context, IN const char *const name)
{
	int ret;
	cl_status_t status;

	CL_ASSERT(p_thread);
	CL_ASSERT(pfn_callback);

	cl_thread_construct(p_thread);

	/* Initialize the thread structure */
	p_thread->osd.task = NULL;
	p_thread->pfn_callback = pfn_callback;
	p_thread->context = context;

	strncpy(p_thread->name, name, 16);
	p_thread->name[15] = '\0';

	init_waitqueue_head(&p_thread->osd.wqueue);

	status = cl_event_init(&p_thread->osd.kill_event, FALSE);
	if (status != CL_SUCCESS) {
		cl_thread_destroy(p_thread);
		return (CL_ERROR);
	}

	ret = kernel_thread(__cl_kernel_system_thread, (void *) p_thread, 0);

	if (ret < 0) {
		cl_thread_destroy(p_thread);
		return (CL_ERROR);
	}

	p_thread->osd.state = CL_INITIALIZED;

	return (CL_SUCCESS);
}

/* Destroy the thread.
 * If this function is called after the callback function terminated causing
 * the thread to terminate, this function returns immediately.
 * If this function is called before the callback function returns, the
 * caller will wait for the kill event forever.
 */
void
cl_thread_destroy(IN cl_thread_t * const p_thread)
{
	CL_ASSERT(cl_is_state_valid(p_thread->osd.state));

	if (p_thread->osd.state == CL_INITIALIZED) {
		/* Wait without time out */
		cl_event_wait_on(&p_thread->osd.kill_event,
				 EVENT_NO_TIMEOUT, FALSE);
	}

	p_thread->pfn_callback = NULL;
	p_thread->context = NULL;

	cl_event_destroy(&p_thread->osd.kill_event);
}

/* 
 * Suspend a thread execution for up to "pause_ms" milliseconds.  The thread may
 * resume prior to "pause_ms" time due to a signal.
 */
void
cl_thread_suspend(IN const uint32_t pause_ms)
{
	unsigned long expiration;

	CL_ASSERT(!in_interrupt());

	/* Time is already in milliseconds. */
	if (!pause_ms) {
		expiration = 1;
	} else {
		expiration = (pause_ms + MISECS_PER_TICK - 1) / MISECS_PER_TICK;
	}

	set_current_state(TASK_INTERRUPTIBLE);
	expiration = schedule_timeout(expiration);

	/* task state has been set to TASK_RUNNING on exit from schedule_timeout() - ps */

}

/* 
 * Delay execution of thread for some number of microseconds.
 * This uses Linuxs' udelay() function. Remember, suggested maximum
 * value of microseconds is 1000, ie, 1 milliseconds for udelay. Therefore
 * we detect the condition and use mdelay which does not have the restriction.
 */
void
cl_thread_stall(IN const uint32_t pause_us)
{
	CL_ASSERT(pause_us);
	if (pause_us > 1000UL)
		mdelay((pause_us / 1000UL) + 1);
	else
		udelay(pause_us);
}

/* Number of processors on this system */
uint32_t
cl_proc_count(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	return (num_booting_cpus());	/* from <linux/smp.h> set to 1 for UP. */
#else
	return (smp_num_cpus);	/* from <linux/smp.h> set to 1 for UP. */
#endif
}

boolean_t
cl_is_current_thread(IN const cl_thread_t * const p_thread)
{
	CL_ASSERT(p_thread->osd.state == CL_INITIALIZED);

	return (current == p_thread->osd.task);
}
