/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	Abstraction of Timer create, destroy functions.
 *
 * Environment
 *	Linux Kernel Mode
 *
 * $Revision: 1.1 $
 */

#include <complib/cl_timer.h>
#include <complib/cl_memory.h>
#include <linux/sched.h>
#include <linux/time.h>

#ifndef	USEC_PER_SEC
#define USEC_PER_SEC		((unsigned long)1000000)
#endif

/*
 * Timer callback function. This function is called by the kernel
 * whenever the timer expires.
 */
static
    void
__cl_timer_callback(IN const unsigned long arg)
{
	cl_timer_t *p_timer = (cl_timer_t *) arg;

	p_timer->timer.expires = (uintn_t) ~ 0;
	p_timer->in_timer_cb = TRUE;
	p_timer->pfn_callback((void *) p_timer->context);
	p_timer->in_timer_cb = FALSE;
}

void
cl_timer_construct(IN cl_timer_t * const p_timer)
{
	cl_memclr(p_timer, sizeof (cl_timer_t));
	p_timer->state = CL_UNINITIALIZED;
	p_timer->in_timer_cb = FALSE;
}

/*
 * Initialize a timer object. Whenver the timer expires, the kernel
 * calls the user supplied "TimerCallback" function with the
 * user supplied context "pContext".
 */
cl_status_t
cl_timer_init(IN cl_timer_t * const p_timer,
	      IN cl_pfn_timer_callback_t pfn_callback,
	      IN const void *const context)
{
	CL_ASSERT(p_timer);

	cl_timer_construct(p_timer);

	/* Only initialize the timer, don't start it */
	init_timer(&p_timer->timer);

	p_timer->timer.data = (unsigned long) p_timer;
	p_timer->timer.function = __cl_timer_callback;

	p_timer->pfn_callback = pfn_callback;
	p_timer->context = context;

	p_timer->timer.expires = (uintn_t) ~ 0;

	/* timer is in "initialized" state now. */
	p_timer->state = CL_INITIALIZED;

	return (CL_SUCCESS);
}

/* Destroy a timer object. Make sure that the time is stopped */
void
cl_timer_destroy(IN cl_timer_t * const p_timer)
{
	CL_ASSERT(p_timer);
	CL_ASSERT(cl_is_state_valid(p_timer->state));

	if ((p_timer->state == CL_INITIALIZED))
		cl_timer_stop(p_timer);

	p_timer->state = CL_UNINITIALIZED;
}

/* Start a timer created previously */
cl_status_t
cl_timer_start(IN cl_timer_t * const p_timer, IN const uint32_t time_ms)
{
	ulong interval;

	CL_ASSERT(p_timer);

	/* the timer must have been initialized */
	CL_ASSERT(p_timer->state == CL_INITIALIZED);

	cl_timer_stop(p_timer);

	/*
	 * Minimum timer resoultion that can be provided with jiffies is
	 * 10ms. If the caller is requesting anything less than that, we
	 * have no choice
	 */
	interval = (time_ms * HZ) / 1000;
	if (!interval)
		interval = 1;

	p_timer->timer.expires = jiffies + interval;

	add_timer(&p_timer->timer);

	return (CL_SUCCESS);
}

/* Start a timer created previously */
cl_status_t
cl_timer_trim(IN cl_timer_t * const p_timer, IN const uint32_t time_ms)
{
	ulong interval;

	CL_ASSERT(p_timer);

	/* the timer must have been initialized */
	CL_ASSERT(p_timer->state == CL_INITIALIZED);

	interval = (time_ms * HZ) / 1000;
	if ((jiffies + interval) >= p_timer->timer.expires)
		return (CL_SUCCESS);

	/* mod_timer will activate the timer if it's not active */
	return (mod_timer(&p_timer->timer,(jiffies + interval)));
}

void
cl_timer_stop(IN cl_timer_t * const p_timer)
{
	CL_ASSERT(p_timer);
	CL_ASSERT(p_timer->state == CL_INITIALIZED);

	/*
	 * If we are already in callbacks, then del_timer_sync will
	 * spin for ever in SMP case.
	 */
	if (p_timer->in_timer_cb == TRUE)
		del_timer(&p_timer->timer);
	else
		del_timer_sync(&p_timer->timer);
	p_timer->timer.expires = (uintn_t) ~ 0;
}

/*
 * Return the number of micro seconds that have passed since the system
 * was booted
 */
uint64_t
cl_get_time_stamp(void)
{
	uint64_t tstamp;
	struct timeval tval;

	do_gettimeofday(&tval);
	tstamp =
	    ((uint64_t) tval.tv_sec * USEC_PER_SEC) + (uint64_t) tval.tv_usec;

	return (tstamp);
}

/*
 * Return the number of seconds that have passed since the system
 * was booted
 */
uint32_t
cl_get_time_stamp_sec(void)
{
	return (jiffies / HZ);
}
