/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

#include <complib/cl_async_proc.h>

#define CL_ASYNC_PROC_MIN		16
#define CL_ASYNC_PROC_GROWSIZE	16

/* Worker function declaration. */
static void
 __cl_async_proc_worker(IN void *const context);

void
cl_async_proc_construct(IN cl_async_proc_t * const p_async_proc)
{
	CL_ASSERT(p_async_proc);

	cl_qlist_init(&p_async_proc->item_queue);
	cl_spinlock_construct(&p_async_proc->lock);
	cl_thread_pool_construct(&p_async_proc->thread_pool);
}

cl_status_t
cl_async_proc_init(IN cl_async_proc_t * const p_async_proc,
		   IN const uint32_t thread_count, IN const char *const name)
{
	cl_status_t status;

	CL_ASSERT(p_async_proc);

	cl_async_proc_construct(p_async_proc);

	status = cl_spinlock_init(&p_async_proc->lock);
	if (status != CL_SUCCESS) {
		cl_async_proc_destroy(p_async_proc);
		return (status);
	}

	status = cl_thread_pool_init(&p_async_proc->thread_pool, thread_count,
				     __cl_async_proc_worker, p_async_proc,
				     name);
	if (status != CL_SUCCESS)
		cl_async_proc_destroy(p_async_proc);

	return (status);
}

void
cl_async_proc_destroy(IN cl_async_proc_t * const p_async_proc)
{
	/* Destroy the thread pool first so that the threads stop. */
	cl_thread_pool_destroy(&p_async_proc->thread_pool);

	/* Flush all queued callbacks. */
	if (!cl_is_qlist_empty(&p_async_proc->item_queue)) {
		__cl_async_proc_worker(p_async_proc);
	}

	/* Destroy the spinlock. */
	cl_spinlock_destroy(&p_async_proc->lock);
}

void
cl_async_proc_queue(IN cl_async_proc_t * const p_async_proc,
		    IN cl_async_proc_item_t * const p_item)
{
	CL_ASSERT(p_async_proc);
	CL_ASSERT(p_item->pfn_callback);

	/* Enqueue this item for processing. */
	cl_spinlock_acquire(&p_async_proc->lock);
	cl_qlist_insert_tail(&p_async_proc->item_queue,
			     &p_item->pool_item.list_item);
	cl_spinlock_release(&p_async_proc->lock);

	/* Signal the thread pool to wake up. */
	cl_thread_pool_signal(&p_async_proc->thread_pool);
}

static void
__cl_async_proc_worker(IN void *const context)
{
	cl_async_proc_t *p_async_proc = (cl_async_proc_t *) context;
	cl_list_item_t *p_list_item;
	cl_async_proc_item_t *p_item;

	/* Process items from the head of the queue until it is empty. */
	cl_spinlock_acquire(&p_async_proc->lock);
	p_list_item = cl_qlist_remove_head(&p_async_proc->item_queue);
	while (p_list_item != cl_qlist_end(&p_async_proc->item_queue)) {
		/* Release the lock during the user's callback. */
		cl_spinlock_release(&p_async_proc->lock);

		/* Invoke the user callback. */
		p_item = (cl_async_proc_item_t *) p_list_item;
		p_item->pfn_callback(p_item);

		/* Acquire the lock again to continue processing. */
		cl_spinlock_acquire(&p_async_proc->lock);
		/* Get the next item in the queue. */
		p_list_item = cl_qlist_remove_head(&p_async_proc->item_queue);
	}

	/* The queue is empty.  Release the lock and return. */
	cl_spinlock_release(&p_async_proc->lock);
}
