/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	System Callback abstractions.
 *
 * Environment:
 *	All non-native implementations.
 *
 * $Revision: 1.1 $
 */

#include <complib/cl_syscallback.h>
#include <complib/cl_threadpool.h>
#include <complib/cl_spinlock.h>
#include <complib/cl_memory.h>
#include <complib/cl_debug.h>

/*
 * Called once from the framework that initializes and loads
 * the Public library
 */
#define SYSCALLBACK_MIN_THREADS		(4)	/* Low water mark */
#define SYSCALLBACK_MAX_THREADS		(10)	/* High water mark */
#define SYSCALLBACK_GROW_COUNT		(10)	/* Grow limit */

typedef struct _cl_sys_callback_obj_t {
	cl_spinlock_t lock;
	cl_qlist_t high_queue;
	cl_qlist_t low_queue;
	cl_thread_pool_t thread_pool;
	cl_qpool_t free_pool;
	uint32_t thread_count;
	cl_state_t state;

} cl_sys_callback_obj_t;

/* Global sytem callback object. */
static cl_sys_callback_obj_t g_sys_cb;

static void
 __cl_sys_callback_wrapper(IN void *const context);

void
__cl_sys_callback_construct(void)
{
	cl_memclr(&g_sys_cb, sizeof (cl_sys_callback_obj_t));

	cl_spinlock_construct(&g_sys_cb.lock);

	cl_thread_pool_construct(&g_sys_cb.thread_pool);

	cl_qpool_construct(&g_sys_cb.free_pool);

	g_sys_cb.state = CL_UNINITIALIZED;
}

/* Initialize the System callback module. */
cl_status_t
__cl_sys_callback_init(void)
{
	cl_status_t status;

	__cl_sys_callback_construct();

	g_sys_cb.thread_count = SYSCALLBACK_MIN_THREADS;

	/* Set the state to initialized.  The cleanup code will reset as needed. */
	g_sys_cb.state = CL_INITIALIZED;

	/* Initialize the state to error to make error cases simpler. */
	cl_qlist_init(&g_sys_cb.high_queue);
	cl_qlist_init(&g_sys_cb.low_queue);

	status = cl_spinlock_init(&g_sys_cb.lock);
	if (status != CL_SUCCESS) {
		__cl_sys_callback_destroy();
		return (status);
	}

	status =
	    cl_thread_pool_init(&g_sys_cb.thread_pool, g_sys_cb.thread_count,
				__cl_sys_callback_wrapper, NULL, "SysCallback");
	if (status != CL_SUCCESS) {
		__cl_sys_callback_destroy();
		return (status);
	}

	status =
	    cl_qpool_init(&g_sys_cb.free_pool, 0, 0, SYSCALLBACK_GROW_COUNT,
			  sizeof (cl_sys_callback_item_t), NULL, NULL, NULL);
	if (status != CL_SUCCESS)
		__cl_sys_callback_destroy();

	return (status);
}

void
__cl_sys_callback_destroy(void)
{
	uint32_t count = 0;

	CL_ASSERT(cl_is_state_valid(g_sys_cb.state));

	/* Destroy the thread pool. This blocks until all threads have exited. */
	cl_thread_pool_destroy(&g_sys_cb.thread_pool);

	/* Cleanup any pending callbacks. */
	if (g_sys_cb.state == CL_INITIALIZED) {
		while (!cl_is_qlist_empty(&g_sys_cb.high_queue)) {
			cl_qpool_put(&g_sys_cb.free_pool, (cl_pool_item_t *)
				     cl_qlist_remove_head(&g_sys_cb.
							  high_queue));
			count++;
		}

		if (count) {
			cl_dbg_out("cl_sys_callback: destroy: %d high priority "
				   "items destroyed.\n", count);
		}

		count = 0;

		while (!cl_is_qlist_empty(&g_sys_cb.low_queue)) {
			cl_qpool_put(&g_sys_cb.free_pool, (cl_pool_item_t *)
				     cl_qlist_remove_head(&g_sys_cb.low_queue));
			count++;
		}

		if (count) {
			cl_dbg_out("cl_sys_callback: destroy: %d low priority "
				   "items destroyed.\n", count);
		}
	}

	cl_qpool_destroy(&g_sys_cb.free_pool);

	cl_spinlock_destroy(&g_sys_cb.lock);

	g_sys_cb.state = CL_UNINITIALIZED;
}

cl_sys_callback_item_t *
cl_sys_callback_get(IN const void *const get_context)
{
	cl_sys_callback_item_t *p_item;

	if (g_sys_cb.state != CL_INITIALIZED)
		return (NULL);

	/* Simple, return the next available one from the free pool */
	cl_spinlock_acquire(&g_sys_cb.lock);
	p_item = (cl_sys_callback_item_t *) cl_qpool_get(&g_sys_cb.free_pool);
	cl_spinlock_release(&g_sys_cb.lock);

	if (!p_item)
		return (NULL);

	p_item->get_context = get_context;
	p_item->pfn_callback = NULL;

	return (p_item);
}

void
cl_sys_callback_put(IN cl_sys_callback_item_t * const p_item)
{
	CL_ASSERT(p_item);
	CL_ASSERT(g_sys_cb.state == CL_INITIALIZED);
	CL_ASSERT(p_item->pfn_callback == NULL);

	cl_spinlock_acquire(&g_sys_cb.lock);
	cl_qpool_put(&g_sys_cb.free_pool, &p_item->pool_item);
	cl_spinlock_release(&g_sys_cb.lock);
}

cl_status_t
cl_sys_callback_queue(IN cl_sys_callback_item_t * const p_item,
		      IN cl_pfn_sys_callback_t pfn_callback,
		      IN const void *const queue_context,
		      IN const boolean_t high_priority)
{
	CL_ASSERT(g_sys_cb.state == CL_INITIALIZED);
	CL_ASSERT(p_item->pfn_callback == NULL);

	p_item->pfn_callback = pfn_callback;
	p_item->queue_context = queue_context;

	/* Insert at tail of the appropriate queue. */
	cl_spinlock_acquire(&g_sys_cb.lock);
	if (high_priority) {
		cl_qlist_insert_tail(&g_sys_cb.high_queue,
				     &p_item->pool_item.list_item);
	} else {
		cl_qlist_insert_tail(&g_sys_cb.low_queue,
				     &p_item->pool_item.list_item);
	}
	cl_spinlock_release(&g_sys_cb.lock);

	/* Signal one of the threads from the thread pool items */
	return (cl_thread_pool_signal(&g_sys_cb.thread_pool));
}

/*
 * This function pulls the next available item off the high
 * or low priority queues
 */
static cl_sys_callback_item_t *
__cl_get_work_item(void)
{
	cl_sys_callback_item_t *p_item = NULL;

	cl_spinlock_acquire(&g_sys_cb.lock);

	/* Check to see if there was anything on the high priority queue. */
	if (!cl_is_qlist_empty(&g_sys_cb.high_queue)) {
		p_item = (cl_sys_callback_item_t *)
		    cl_qlist_remove_head(&g_sys_cb.high_queue);
	} else {
		/*
		 * Nothing on the high priority queue.  Check the low priority
		 * queue.
		 */
		if (!cl_is_qlist_empty(&g_sys_cb.low_queue)) {
			p_item = (cl_sys_callback_item_t *)
			    cl_qlist_remove_head(&g_sys_cb.low_queue);
		}
	}
	cl_spinlock_release(&g_sys_cb.lock);

	return (p_item);
}

/*
 * This is our SysCallback Thread function.
 * It looks at the high priority queue first, if there is some work to do
 * pick it up and run it. Else move on to the low priority queue
 */
static void
__cl_sys_callback_wrapper(IN void *const context)
{
	cl_sys_callback_item_t *p_item;
	cl_pfn_sys_callback_t pfn_callback;

	p_item = __cl_get_work_item();

	while (p_item) {
		CL_ASSERT(p_item->pfn_callback);
		/* Invoke the user callback. */
		pfn_callback = p_item->pfn_callback;
		/*
		 * Clear the item's callback to allow the callback recipient to
		 * put/queue it again.
		 */
		p_item->pfn_callback = NULL;
		pfn_callback((void *) p_item->get_context,
			     (void *) p_item->queue_context);
		p_item = __cl_get_work_item();
	}
}
