/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

#include <complib/cl_event.h>
#include <complib/cl_debug.h>
#include <sys/time.h>
#include <sys/errno.h>


void
cl_event_construct(
	IN cl_event_t *p_event )
{
	CL_ASSERT( p_event );

	cl_spinlock_construct( &p_event->spinlock );
	p_event->state = CL_UNINITIALIZED;
}


cl_status_t
cl_event_init(
	IN	cl_event_t* const	p_event,
	IN	const boolean_t		manual_reset )
{
	CL_ASSERT( p_event );

	cl_event_construct( p_event );

	pthread_cond_init( &p_event->condvar, NULL );
	cl_spinlock_init( &p_event->spinlock );
	p_event->signaled = FALSE;
	p_event->manual_reset = manual_reset;
	p_event->state = CL_INITIALIZED;

	return( CL_SUCCESS );
}


void
cl_event_destroy(
	IN	cl_event_t* const	p_event )
{
	CL_ASSERT( cl_is_state_valid( p_event->state ) );

	/* Destroy only if the event was constructed */
	if( p_event->state == CL_INITIALIZED )
	{
		pthread_cond_broadcast( &p_event->condvar );
		pthread_cond_destroy( &p_event->condvar );
	}

	cl_spinlock_destroy( &p_event->spinlock );
	p_event->state = CL_UNINITIALIZED;
}


cl_status_t
cl_event_signal(
	IN	cl_event_t* const	p_event )
{
	/* Make sure that the event was started */
	CL_ASSERT( p_event->state == CL_INITIALIZED );

	cl_spinlock_acquire( &p_event->spinlock );
	p_event->signaled = TRUE;
	/* Wake up one or all depending on whether the event is auto-reseting. */
	if( p_event->manual_reset )
		pthread_cond_broadcast( &p_event->condvar );
	else
		pthread_cond_signal( &p_event->condvar );

	cl_spinlock_release( &p_event->spinlock );

	return( CL_SUCCESS );
}


cl_status_t
cl_event_reset(
	IN	cl_event_t* const	p_event )
{
	/* Make sure that the event was started */
	CL_ASSERT( p_event->state == CL_INITIALIZED );

	cl_spinlock_acquire( &p_event->spinlock );
	p_event->signaled = FALSE;
	cl_spinlock_release( &p_event->spinlock );

	return( CL_SUCCESS );
}


cl_status_t
cl_event_wait_on(
	IN	cl_event_t* const	p_event,
	IN	const uint32_t		wait_us,
	IN	const boolean_t		interruptible )
{
	cl_status_t		status;
	int				wait_ret;
	struct timespec	timeout;
	struct timeval	curtime;

	/* Make sure that the event was Started */
	CL_ASSERT( p_event->state == CL_INITIALIZED );

	cl_spinlock_acquire( &p_event->spinlock );

	/* Return immediately if the event is signalled. */
	if( p_event->signaled )
	{
		if( !p_event->manual_reset )
			p_event->signaled = FALSE;

		cl_spinlock_release( &p_event->spinlock );
		return( CL_SUCCESS );
	}

	/* If just testing the state, return CL_TIMEOUT. */
	if( wait_us == 0 )
	{
		cl_spinlock_release( &p_event->spinlock );
		return( CL_TIMEOUT );
	}

	if( wait_us == EVENT_NO_TIMEOUT )
	{
		/* Wait for condition variable to be signaled or broadcast. */
		if( pthread_cond_wait( &p_event->condvar, &p_event->spinlock.mutex ) )
			status = CL_NOT_DONE;
		else
			status = CL_SUCCESS;
	}
	else
	{
		/* Get the current time */
		if( gettimeofday( &curtime, NULL ) == 0 )
		{
			timeout.tv_sec = curtime.tv_sec + (wait_us / 1000000);
			timeout.tv_nsec = (curtime.tv_usec + (wait_us % 1000000)) * 1000;

			wait_ret = pthread_cond_timedwait( &p_event->condvar,
				&p_event->spinlock.mutex, &timeout );
			if( wait_ret == 0 )
				status = (p_event->signaled ? CL_SUCCESS : CL_NOT_DONE);
			else if( wait_ret == ETIMEDOUT )
				status = CL_TIMEOUT;
			else
				status = CL_NOT_DONE;
		}
		else
		{
			status = CL_ERROR;
		}
	}
	if( !p_event->manual_reset )
		p_event->signaled = FALSE;

	cl_spinlock_release( &p_event->spinlock );
	return( status );
}
