/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

#include <stdio.h>
#include <unistd.h>

#ifdef LINUX
#include <sys/sysinfo.h>
#else
#include <sys/param.h>
#include <sys/sysctl.h>

#endif
#include <complib/cl_thread.h>


/*
 * Internal function to run a new user mode thread.
 * This function is always run as a result of creation a new user mode thread.
 * Its main job is to synchronize the creation and running of the new thread.
 */
void*
__cl_thread_wrapper( void* arg )
{
	cl_thread_t		*p_thread = (cl_thread_t*)arg;

	CL_ASSERT( p_thread );
	CL_ASSERT( p_thread->pfn_callback );

	p_thread->pfn_callback( (void*)p_thread->context );

	return( NULL );
}


void
cl_thread_construct(
	IN	cl_thread_t* const	p_thread )
{
	CL_ASSERT( p_thread );

	p_thread->osd.state = CL_UNINITIALIZED;
}


cl_status_t
cl_thread_init(
	IN	cl_thread_t* const			p_thread,
	IN	cl_pfn_thread_callback_t	pfn_callback,
	IN	const void* const			context,
	IN	const char* const			name )
{
	int				ret;

	CL_ASSERT( p_thread );

	cl_thread_construct( p_thread );

	/* Initialize the thread structure */
	p_thread->pfn_callback = pfn_callback;
	p_thread->context = context;

	ret = pthread_create( &p_thread->osd.id, NULL,
		__cl_thread_wrapper, (void*)p_thread );

	if( ret != 0 )	/* pthread_create returns a "0" for success */
		return( CL_ERROR );

	p_thread->osd.state = CL_INITIALIZED;

	return( CL_SUCCESS );
}


void
cl_thread_destroy(
	IN	cl_thread_t* const	p_thread )
{
	CL_ASSERT( p_thread );
	CL_ASSERT( cl_is_state_valid( p_thread->osd.state ) );

	if( p_thread->osd.state == CL_INITIALIZED )
		pthread_join( p_thread->osd.id, NULL );

	p_thread->osd.state = CL_UNINITIALIZED;
}


void
cl_thread_suspend(
	IN	const uint32_t	pause_ms )
{
	/* Convert to micro seconds */
	usleep( pause_ms * 1000 );
}


void
cl_thread_stall(
	IN	const uint32_t	pause_us )
{
	/*
	 * Not quite a busy wait, but Linux is lacking in terms of high
	 * resolution time stamp information in user mode.
	 */
	usleep( pause_us );
}


uint32_t
cl_proc_count( void )
{
#ifdef LINUX
	uint32_t ret;

	ret = get_nprocs();
	if( !ret)
		return 1;/* Workaround for PPC where get_nprocs() returns 0 */

	return ret;
#else
   int mib[2], nproc;
   size_t len;

   mib[0] = CTL_KERN;
   mib[1] = KERN_NPROCS;
   len = sizeof(nproc);
   sysctl(mib, 2, &nproc, &len, NULL, 0);

   return nproc;
#endif
}


boolean_t
cl_is_current_thread(
	IN	const cl_thread_t* const	p_thread )
{
	pthread_t	current;

	CL_ASSERT( p_thread );
	CL_ASSERT( p_thread->osd.state == CL_INITIALIZED );

	current = pthread_self();
	return( pthread_equal( current, p_thread->osd.id ) );
}
